/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to you under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.jorphan.collections;

import java.util.Map;
import java.util.function.BiConsumer;
import java.util.function.Function;
import java.util.function.Supplier;

public class CopyOnWriteMap<T, K, V> {
    private static final Object DELETED = new Object();

    private final Supplier<? extends Map<K, V>> mapFactory;
    private final Function<? super V, ? extends V> whenUnshared;
    private final Function<? super T, Object> field1Get;
    private final Function<? super T, Object> field2Get;
    @SuppressWarnings("unused")
    private final BiConsumer<T, Object> field1Set;
    private final BiConsumer<? super T, Object> field2Set;

    public CopyOnWriteMap(Supplier<? extends Map<K, V>> mapFactory,
                          Function<? super V, ? extends V> whenUnshared, Function<? super T, Object> field1Get, BiConsumer<T, Object> field1Set,
                          Function<? super T, Object> field2Get, BiConsumer<? super T, Object> field2Set) {
        this.mapFactory = mapFactory;
        this.whenUnshared = whenUnshared;
        this.field1Get = field1Get;
        this.field2Get = field2Get;
        this.field1Set = field1Set;
        this.field2Set = field2Set;
    }

    private Map<K, V> sharedParams(T that) {
        //noinspection unchecked
        return (Map<K, V>) field1Get.apply(that);
    }

    private Map<K, V> modifiedParams(T that) {
        //noinspection unchecked
        return (Map<K, V>) field2Get.apply(that);
    }

    private Map<K, V> createModifiedParams(T that) {
        Map<K, V> map = modifiedParams(that);
        if (map == null) {
            map = mapFactory.get();
            field2Set.accept(that, map);
        }
        return map;
    }

    private static<T> T mask(T v) {
        //noinspection unchecked
        return v == null ? (T) DELETED : v;
    }

    private static<T> T unmask(T v) {
        return v == DELETED ? null : v;
    }

    public V set(T that, K key, V value) {
        Map<K, V> modified = createModifiedParams(that);
        V prev = modified.put(key, mask(value));
        if (prev != null) {
            return unmask(prev);
        }
        // return prev value
        Map<K, V> shared = sharedParams(that);
        return shared != null ? shared.get(key) : null;
    }

    public V get(T that, K key) {
        Map<K, V> modified = modifiedParams(that);
        if (modified != null) {
            V v = modified.get(key);
            if (v != null) {
                return unmask(v);
            }
        }
        Map<K, V> shared = sharedParams(that);
        if (shared == null) {
            return null;
        }
        V v = shared.get(key);
        if (v == null) {
            return null;
        }
        if (modified == null) {
            modified = createModifiedParams(that);
        }
        v = whenUnshared.apply(v);
        modified.put(key, v);
        return v;
    }

    public V getReadOnly(T that, K key) {
        Map<K, V> modified = modifiedParams(that);
        if (modified != null) {
            V v = modified.get(key);
            if (v != null) {
                return unmask(v);
            }
        }
        Map<K, V> shared = sharedParams(that);
        return shared == null ? null : shared.get(key);
    }
}
